#ifndef _rtp_MAT_HH
#define _rtp_MAT_HH


#include "rtp/namespace.hh"
#include "rtp/vec.hh"

#include <assert.h>
#include <math.h> /* sqrt() */


RTP_NAMESPACE_BEGIN

////////////////////////////////////////////////////////////////////////////

#define _RTP_B(i,n,X) { const int i = n; X; }

#ifndef RTP_MAT_UNROLL
#define RTP_MAT_UNROLL 1
#endif

#if RTP_MAT_UNROLL
#define _RTP_MAT_ROW(i,N,X) \
    switch (N) { \
      case 1: { const unsigned int i = 0; X; } break; \
      case 2: { const unsigned int i = 0; X; } \
              { const unsigned int i = 1; X; } break; \
      case 3: { const unsigned int i = 0; X; } \
              { const unsigned int i = 1; X; } \
              { const unsigned int i = 2; X; } break;\
      case 4: { const unsigned int i = 0; X; } \
              { const unsigned int i = 1; X; } \
              { const unsigned int i = 2; X; } \
              { const unsigned int i = 3; X; } break;\
      default: for ( unsigned int i = 0; i < (N); ++ i ) { X; }; \
    }
#else
#define _RTP_MAT_ROW(i,N,X) for ( unsigned int i = 0; i < (N); ++ i ) { X; }
#endif


#if RTP_MAT_UNROLL
#define _RTP_MAT_ROW_PROD(i,N,X,RESULT,OUTER,L,INNER,R) \
    switch (N) { \
      case 1: { const unsigned int i = 0; RESULT = L INNER R; } break; \
      case 2: { const unsigned int i = 0; RESULT = (L INNER R) OUTER; } \
              { const unsigned int i = 1; ; } break; \
      case 3: { const unsigned int i = 0; X; } \
              { const unsigned int i = 1; X; } \
              { const unsigned int i = 2; X; } break;\
      case 4: { const unsigned int i = 0; X; } \
              { const unsigned int i = 1; X; } \
              { const unsigned int i = 2; X; } \
              { const unsigned int i = 3; X; } break;\
      default: for ( unsigned int i = 0; i < (N); ++ i ) { X; }; \
    }
#else
#define _RTP_MAT_ROW_PROD(i,N,X,RESULT,OUTER,L,INNER,R) for ( unsigned int i = 0; i < (N); ++ i ) { X; }
#endif

#define _RTP_MAT_LOOP(i,j,X) \
    _RTP_MAT_ROW(i,M,_RTP_MAT_ROW(j,N,X))


////////////////////////////////////////////////////////////////////////////


template <class T, int M, int N = M> 
class mat
{
public:
  typedef mat<T,M,N> this_t;

private:
  T _v[M][N];

  void init(const T &x)
  {
    _RTP_MAT_LOOP(i,j,_v[i][j] = (T) 0);
    _RTP_MAT_ROW(i,N < M ? N : M, _v[i][i] = x);
  }

  void copy(const T *src)
  {
    _RTP_MAT_LOOP(i,j,_v[i][j] = *(src ++));
  }

public:
  mat<T,M,N>() { };

  mat<T,M,N>(const T &x) { init(x); }

  mat<T,M,N>(const T *x) { init(x); }
  mat<T,M,N>(const mat<T,M,N> &x) { copy(x._v[0]); }

  ~mat<T,M,N>() { }

  T* operator[](unsigned int i) 
  { 
    assert(i <  N); 
    return _v[i];
  }
  const T *operator[](unsigned int i) const 
  { 
    assert(i <  N); 
    return _v[i];
  }


  // addition: Matrix + Matrix => Matrix
  this_t &operator +=(const this_t &x) //@ Addition.
  {
    _RTP_MAT_LOOP(i,j,_v[i][j] += x._v[i][j]);
    return *this;
  }
  this_t operator +(const this_t &x) const 
  { this_t tmp(*this); return tmp += x; }


  // subtraction: Matrix - Matrix => Matrix
  this_t &operator -=(const this_t &x)
  { 
    _RTP_MAT_LOOP(i,j,_v[i][j] -= x._v[i][j]);
    return *this;
  }
  this_t operator -(const this_t &x) const 
  { this_t tmp(*this); return tmp -= x; }


  // scale: Matrix * Scalar => Matrix
  this_t &operator *=(const T &x)
  { 
    _RTP_MAT_LOOP(i,j,_v[i][j] *= x);
    return *this;
  }
  this_t operator *(const T &x) const 
  { this_t tmp(*this); return tmp *= x; }


  // negation: - Matrix => Matrix
  this_t &negate()
  {
    _RTP_MAT_LOOP(i,j,_v[i][j] = - _v[i][j]);
    return *this;
  }
  this_t operator -() const //@ Negation.
  { this_t tmp(*this); return tmp.negate(); }

  ///////////////////////////////////////////////////////
  // Inverse
  //

  this_t &invert()
  {
    assert(N == M);
    // IMPLEMENT;
    return *this;
  }
  
  ///////////////////////////////////////////////////////
  // Transpose

  // template <T> void swap(T &x, T &y)
  // { T t(x); x = y; y = t; }

  mat<T,M,N> &trans() 
  {
    // Must be square.
    assert(N == M);

    // *, A
    // a, *
    if ( N > 1 ) {
      swap(_rep[1][0], _rep[0][1]);
    }

    // *, A, B
    // a, *, C
    // b, C, *
    if ( N > 2 ) {
      swap(_rep[2][0], _rep[0][2]);
      swap(_rep[2][1], _rep[1][2]);
    }

    // *, A, B, D
    // a, *, C, E
    // b, C, *, F
    // d, e, f, *
    if ( N > 3 ) {
      swap(_rep[3][0], _rep[0][3]);
      swap(_rep[3][1], _rep[1][3]);
      swap(_rep[3][2], _rep[2][3]);
    }

    // *, A, B, D, G
    // a, *, C, E, H
    // b, C, *, F, I
    // d, e, f, *, J
    // g, h, i, j, *
    if ( N > 4 ) {
      for ( unsigned int i = 4; i < N; ++ i ) {
	for ( unsigned int j = 0; j < M; ++ j ) {
	  swap(_rep[i][j], _rep[j][i]);
	}
      }
    }

    return *this;
  }

  mat<T,M,N> transpose()
  { 
    if ( N == M ) {
      // Use fast inline.
      mat<T,M,N> t(*this); 
      return t.trans();
    } else {
      // Non-square fallback.
      mat<T,M,N> t; 
      for ( unsigned int i = 0; i < N; ++ i) {
	for ( unsigned int j = 0; j < M; ++ j ) {
	  t._rep[j][i] = _rep[i][j];
	}
      }
      return t;
    }
  }

};


template <class T, int N, int M>
mat<T,M,N> operator * ( const T &x, const mat<T,M,N> &v )
{
  return v * x;
}


/* A[1][M] * B[M][N] = C[1][N] */
template <class T, int N, int M>
vec<T,N> operator * (const vec<T,M> &v, const mat<T,M,N> &m)
{
  vec<T,M> tmp;

  for ( unsigned int i = 0; i < N; ++ i ) {
    tmp[i] = v[0] * m[0][i];
    for ( unsigned int j = 1; j < M; ++ j ) {
      tmp[i] += v[j] * m[j][i];
    }
  }

  return tmp;
}


/* A[M][N] * B[N][1] = C[M][1] */
template <class T, int N, int M>
vec<T,M> operator * (const mat<T,M,N> &m, const vec<T,N> &v)
{
  vec<T,M> tmp;

  for ( unsigned int i = 0; i < M; ++ i ) {
    tmp[i] = v[0] * m[i][0];
    for ( unsigned int j = 1; j < N; ++ j ) {
      tmp[i] += v[j] * m[i][j];
    }
  }

  return tmp;
}

/* A[M][N] * B[N][L] = C[M][L] */
template <class T, int N, int M, int L>
mat<T,M,L> operator * (const mat<T,M,N> &a, const mat<T,N,L> &b)
{
  mat<T,M,L> c;

  for ( unsigned int i = 0; i < M; ++ i ) {
    for ( unsigned int j = 0; j < L; ++ j ) {
      c[i][j] = a[i][0] * b[0][j];
      for ( unsigned int k = 1; k < N; ++ k ) {
	c[i][j] += a[i][k] * b[k][j];
      }
    }
  }

  return c;
}


/////////////////////////////////////////////////////////////////////


template <class OS, class T, int M, int N> 
OS &operator <<(OS &os, const mat<T,M,N> &x)
{
  os << "{\n";
  _RTP_MAT_LOOP(i,j,os << x[i][j] << (j == M - 1 ? ",\n" : ", "));
  os << "}";

  return os;
}


template <class OS, class T, int M, int N> 
OS &operator >>(OS &os, mat<T,M,N> &x)
{
  char c;

  os >> c;
  _RTP_MAT_LOOP(i,j,((os >> c >> x[i][j])));
  os >> c;

  return os;
}

RTP_NAMESPACE_END

#endif
