#ifndef _rtp_ORIENTATION_HH
#define _rtp_ORIENTATION_HH

#include "rtp/namespace.hh"

RTP_NAMESPACE_BEGIN

template <class T, int N> class orientation;

RTP_NAMESPACE_END

#include "rtp/frame.hh"
#include "rtp/vector.hh"
#include "rtp/location.hh"
#include "rtp/vec.hh"
#include <math.h> /* cos(), sin() */

RTP_NAMESPACE_BEGIN

template <class T, int N> 
class orientation
{
public:
  static const int rep_N = (N - 1) * 2;

  typedef orientation<T,N> this_t;
  typedef vector<T,rep_N> rep_t;
  typedef frame<T,N> frame_t;
  typedef vector<T,N> vector_t;
  typedef location<T,N> location_t;

private:
  rep_t _v;
  
  // N = 2: a two dimensional orientation is a planar rotation
  // _v[0] contains the x component of the rotated x axis
  // _v[1] contains the y component of the rotated x axis

  // N = 3: a three dimensional orientation is a 3-space rotation,
  // represented as a quaternion which is a 4-tuple.
  
public:

  const rep_t & rep() const { return _v; }
  const rep_t & rep(const frame_t *f) const 
  { 
    this_t tmp(*this);
    to_frame(0, tmp, f);
    return tmp._v;
  }

  explicit orientation<T,N>(const rep_t &x, const frame_t *f = 0) 
    : _v(x) 
  { to_frame(f, *this);  }
 
  // As expressed as a 2-D rotation angle.
  explicit orientation<T,N>(const T &x1, const frame_t *f = 0)
  {
    assert(N == 2);
    _v[0] = cos(x1);
    _v[1] = sin(x1);
    to_frame(f, *this);
  }

  // As expressed as a 2-D X axis.
  explicit orientation<T,N>(const T &x1, const T &x2, const frame_t *f = 0) 
    : _v(x1, x2) 
  { 
    assert(N == 2);
    _v.normalize();
    to_frame(f, *this);
  }

  // As expressed as a quaternion.
  explicit orientation<T,N>(const T &x1, const T &x2, const T &x3, const T &x4, const frame_t *f = 0) 
    : _v(x1, x2, x3, x4) 
  { 
    assert(N == 3);
    to_frame(f, *this);
  }

  // As expressed as an axis and angle
  explicit orientation<T,N>(const vector<T,N> &axis, const T &angle)
  {
    // IMPLEMENT.
    to_frame(f, *this);
  }

  /////////////////////////////////////////////////////
  // Transform:
  // vector
  // location

  ////////////////////////////////////////////////////
  // forward transform.
  //

  template <class X> X operator *(const X &x) const // forward transform.
  {
    if ( N == 2 ) {
      rep_t _y = ~ _v;
      // _y
      // ^
      // |
      // |
      // +----> _v
      return X(
			 x[0] * _v[0] + x[1] * _y[0], 
			 x[0] * _v[1] + x[1] * _y[1]
			 );
    }
    return X();
  }

  /////////////////////////////////////////////////////
  // Inverse transform
  //

  template <class X> X operator /(const X &x) const
  {
    if ( N == 2 ) {
      rep_t _y = ~ _v;
      return X(
	       x[0] * _v[0] - x[1] * _y[0],
	       x[1] * _y[1] - x[0] * _v[1] 
	       );      
    }
    return X();
  }


  /////////////////////////////////////////////////////
  // Concatenation.
  //


  this_t &operator *=(const this_t &o)
  {
    _v = o * _v;
    _v.normalize();
    return *this;
  }

  this_t operator *(const this_t &o)
  {
    this_t tmp(*this);
    return tmp *= o;
  }

  this_t &operator /=(const this_t &o)
  {
    _v = o / _v;
    _v.normalize();
    return *this;
  }

  this_t operator /(const this_t &o)
  {
    this_t tmp(*this);
    return tmp /= o;
  }

};


template <class X, class T, int N>
X operator *(const X &x, const orientation<T,N> &o)
{
  return o * x;
}


template <class X, class T, int N>
X operator /(const X &x, const orientation<T,N> &o)
{
  return o / x;
}


template <class OS, class T, int N> 
OS &operator <<(OS &os, const orientation<T,N> &x)
{
  os << "orientation " << x.rep();

  return os;
}


RTP_NAMESPACE_END

#endif
