#ifndef _rtp_TENSOR2_HH
#define _rtp_TENSOR2_HH


#include "rtp/namespace.hh"

RTP_NAMESPACE_BEGIN

template <class T, int N> class tensor2;

RTP_NAMESPACE_END

#include "rtp/frame.hh"
#include "rtp/vector.hh"
#include "rtp/basis.hh"
#include "rtp/mat.hh"

RTP_NAMESPACE_BEGIN

template <class T, int N> 
class tensor2
{
public:
  typedef tensor2<T,N> this_t;
  typedef mat<T,N,N> rep_t;
  typedef frame<T,N> frame_t;
  typedef vector<T,N> vector_t;
  typedef basis<T,N> basis_t;

private:
  rep_t _rep;

public:

  tensor2<T,N>()
  { }

  explicit tensor2<T,N>(const T &x, const frame_t *f = 0) 
    : _rep(x) 
  { to_frame(f, *this); }

  tensor2<T,N>(const T *x, const frame_t *f = 0) 
    : _rep(x)
  { to_frame(f, *this); }

  tensor2<T,N>(const rep_t &x, const frame_t *f = 0) 
    : _rep(x)
  { to_frame(f, *this); }

  tensor2<T,N>(const T &x1, const T &x2, const frame<T,N> *f = 0) 
    : _rep(x1, x2) 
  { to_frame(f, *this); }

  tensor2<T,N>(const T &x1, const T &x2, const T &x3, const frame<T,N> *f = 0) 
    : _rep(x1, x2, x3) 
  { to_frame(f, *this); }

  T *operator[](int i) { return _rep[i]; };
  const T *operator[](int i) const { return _rep[i]; };

  const rep_t &rep() const { return _rep; }
  rep_t rep(const frame_t *f) const
  {
    this_t tmp(*this);
    to_frame((frame_t*) 0, tmp, f);
    return tmp.rep();
  }

  // Tensor2 * Scalar => Tensor2
  this_t &operator *=(const T &x)
  { _rep *= x; return *this; }
  this_t operator *(const T &x)
  { this_t t(*this); return t *= x; }

  // Tensor2 * Vector => Vector
  vector_t    operator *  (const vector_t &x) const
  { return vector_t(_rep * x.rep()); }

  // Tensor2 + Tensor2 => Tensor2
  this_t &operator +=(const this_t &x)
  { _rep += x._rep; return *this; }
  this_t operator +(const this_t &x) const
  { this_t t(*this); return t += x; }

  // Tensor2 - Tensor2 => Tensor2
  this_t &operator -=(const this_t &x)
  { _rep -= x._rep; return *this; }
  this_t operator -(const this_t &x) const
  { this_t t(*this); return t += x; }

  // Tensor2 * Tensor2 => Tensor2
  this_t &operator *=(const this_t &x)
  { _rep *= x._rep; return *this; }
  this_t operator *(const this_t &x) const
  { this_t t(*this); return t *= x; }

  // Tensor2 * Basis => Basis
  basis_t operator *(const basis_t &x) const
  { return basis_t(_rep * x.rep()); }

  // negation: - Tensor2 => Tensor2
  this_t &negate()
  { _rep.negate(); return *this; }
  this_t operator -() const
  { this_t t(*this); t.negate(); return t; }

  // transpose: Tensor2^-1 => Tensor2
  this_t &trans()
  {  _rep.trans(); return *this; }
  this_t transpose() const
  { this_t t(*this); t.trans(); return t; }
  this_t operator ~() const
  { return *this.transpose(); }

  // inverse: Tensor2^-1 => Tensor2
  this_t &invert()
  {  _rep.invert(); return *this; }
  this_t inverse() const
  { this_t t(*this); t.invert(); return t; }
  
};

// scalar: Scalar * Tensor2 => Tensor2
template <class T> tensor2<T,N> operator *(const T &s, const tensor2<T,N> &v)
{ 
  return v * s;
}


// mulitplication: Basis * Tensor2 => Basis
template <class T> basis<T,N> operator *(const basis<T,N> &s, const tensor2<T,N> &v)
{ 
  return basis<T,N>(v.rep() * s.rep());
}


template <class OS, class T, int N> 
OS &operator <<(OS &os, const tensor2<T,N> &x)
{
  os << "tensor " << x.rep();

  return os;
}


RTP_NAMESPACE_END


#endif
