#ifndef _rtp_VECTOR_HH
#define _rtp_VECTOR_HH


#include "rtp/namespace.hh"

RTP_NAMESPACE_BEGIN

template <class T, int N> class vector;

RTP_NAMESPACE_END

#include "rtp/frame.hh"
#include "rtp/location.hh"
#include "rtp/tensor2.hh"
#include "rtp/basis.hh"

#include "rtp/vec.hh"

RTP_NAMESPACE_BEGIN

template <class T, int N> 
class vector
{
public:
  typedef vector<T,N> this_t;
  typedef vec<T,N> rep_t;
  typedef frame<T,N> frame_t;
  typedef location<T,N> location_t;
  typedef tensor2<T,N> tensor2_t;
  typedef basis<T,N> basis_t;

private:
  rep_t _rep;
  // vec<T,N> _rep;

public:

  vector<T,N>()
  { }

  explicit vector<T,N>(const T &x, const frame_t *f = 0) 
    : _rep(x) 
  { to_frame(f, *this); }

  explicit vector<T,N>(const T *x, const frame_t *f = 0) 
    : _rep(x)
  { to_frame(f, *this); }

  explicit vector<T,N>(const rep_t &x, const frame_t *f = 0) 
    : _rep(x)
  { to_frame(f, *this); }

  explicit vector<T,N>(const T &x1, const T &x2, const frame<T,N> *f = 0) 
    : _rep(x1, x2) 
  { to_frame(f, *this); }

  explicit vector<T,N>(const T &x1, const T &x2, const T &x3, const frame<T,N> *f = 0) 
    : _rep(x1, x2, x3) 
  { to_frame(f, *this); }

  T &operator[](int i) { return _rep[i]; };
  const T &operator[](int i) const { return _rep[i]; };

  const rep_t &rep() const { return _rep; }
  rep_t rep(const frame_t *f) const
  {
    this_t tmp(*this);
    to_frame((frame_t*) 0, tmp, f);
    return tmp.rep();
  }

  // addition: Vector + Vector => Vector
  this_t&   operator += (const this_t& x)
  { _rep += x._rep; return *this; }
  this_t    operator +  (const this_t &x) const
  { this_t tmp(*this); tmp._rep += x._rep; return tmp; }

  // subtraction: Vector - Vector => Vector
  this_t &operator -= (const this_t &x)   //@ Subtraction.
  { _rep -= x._rep; return *this; }
  this_t    operator -  (const this_t &x) const  //@ Subtraction.
  { this_t tmp(*this); tmp._rep -= x._rep; return tmp; }
  
  // scale: Vector * Scalar => Vector
  this_t &operator *= (const T &x)
  { _rep *= x; return *this; }
  this_t  operator *  (const T &x) const
  { this_t tmp(*this); tmp._rep *= x; return tmp; }

  // inner product: Vector . Vector => Scalar
  T operator % (const this_t &x)
  { return _rep * x._rep; }

  // cross product: Vector ^ Vector => Vector
  this_t operator ^(const this_t &x) const
  { return (this_t)(_rep ^ x._rep); }

  // outer product: Vector * Vector => Vector
  this_t operator *(const this_t &x) const
  { return (this_t)(_rep * x._rep); }
  
  // displacement: Vector + Location => Location 
  location_t  operator + (const location_t &x) const //@ Displacement: vector + location = location
  { return x + *this; }

  // cross product: Vector ^ Tensor2 => Tensor2
  tensor2_t operator ^(const tensor2_t &x) const
  { return (tensor2_t)(_rep ^ x._rep); }

  // cross product: Vector ^ Basis => Basis
  basis_t operator ^(const basis_t &x) const
  { return (basis_t)(_rep ^ x._rep); }

  // Norm.
  this_t &normalize()
  { _rep.normalize(); return *this; }
  T norm() const
  { return _rep.norm(); }

  // Normal vector: | Vector | => Vector
  this_t operator !() const
  { return this_t(! _rep); }

  // Orthogonal vector: Vector => Vector
  this_t operator ~() const
  {  return this_t(~ _rep); }

};


// Scalar * Vector => Vector
template <class T, int N> vector<T,N> operator *(const T &s, const vector<T,N> &v)
{ 
  return v * s;
}


template <class OS, class T, int N> 
OS &operator <<(OS &os, const vector<T,N> &x)
{
  os << "vector " << x.rep();

  return os;
}


RTP_NAMESPACE_END


#endif
